/*
	IntelHEX`(hex.c)
*/

#include <stdio.h>
#include <string.h>
#include "hex.h"

#define FALSE	0
#define TRUE	1

typedef struct {
	char colon;          /* R[hJnL ":" */
	char length[2];      /* f[^             */
	char offset[4];      /* ItZbgAhX   */
	char record_type[2]; /* R[h^Cv       */
	char data[256];      /* f[^               */
} Intelhex;

/*
	𐔒l(1byte)ɕϊ
*/
static int h2i(char c)
{
	switch(c) {
	case '0':
		return 0;
	case '1':
		return 1;
	case '2':
		return 2;
	case '3':
		return 3;
	case '4':
		return 4;
	case '5':
		return 5;
	case '6':
		return 6;
	case '7':
		return 7;
	case '8':
		return 8;
	case '9':
		return 9;
	case 'A':
	case 'a':
		return 0xa;
	case 'B':
	case 'b':
		return 0xb;
	case 'C':
	case 'c':
		return 0xc;
	case 'D':
	case 'd':
		return 0xd;
	case 'E':
	case 'e':
		return 0xe;
	case 'F':
	case 'f':
		return 0xf;
	default:
		return -16;
	}
}
static int hex2i(const char *hex)
{
	int val;

	if((val = (h2i(*hex) << 4) | h2i(*(hex + 1))) < 0)
		return -1;
	return val;
}

/*
	IntelHEX`̃eLXgoCiɕϊ
*/
static off_t hex2bin(void *mem, off_t *offset, const char *hex, size_t mem_size, int check)
{
	const Intelhex *h = (const Intelhex *)hex;
	int val, sum;
	off_t len;
	const char *q;
	unsigned char *p, *w, *last = (char *)mem + mem_size;

	*offset = 0;

	/* R[h擪L */
	if(h->colon != ':')
		return -1;
	
	/* R[h^Cv */
	switch(hex2i(h->record_type)) {
	case 0x00:	/* f[^R[h */
		break;
	case 0x01:	/* GhR[h */
		return 0;
	default:	/* ̑̃R[h(Ή) */
		return -1;
	}
	sum = 0;

	/* f[^ */
	len = hex2i(h->length);
	if(len < 0)
		return -1;
	if(len == 0)
		return 0;
	sum += (len & 0xff);

	/* ItZbgAhX */
	*offset = hex2i(h->offset) << 8 | hex2i(&h->offset[2]);
	if(*offset < 0)
		return -1;
	sum += ((*offset & 0xff00) >> 8) + (*offset & 0x00ff);

	/* f[^ */
	if(mem == NULL && !check)
		return len;
	w = mem + *offset;
	for(p = w, q = h->data; p != w + len; p++, q += 2) {
		if((val = hex2i(q)) < 0)
			return -1;
		sum += val;
		if(mem != NULL && p < last)
			*p = val;
	}
	sum = (~sum + 1) & 0xff;

	/* `FbNT */
	if(check && hex2i(q) != sum)
		return -1;

	return len;
}

/*
	l(1byte)𕶎ɕϊ
*/
static char *int2c(char *hex, int i)
{
	const static char int_to_ascii[] = {
		'0', '1', '2', '3', '4', '5', '6', '7',
		'8', '9', 'A', 'B', 'C', 'D', 'E', 'F'
	};

	hex[0] = int_to_ascii[i >> 4];
	hex[1] = int_to_ascii[i & 0xf];
	return hex;
}

/*
	oCiIntelHEX`̃eLXgɕϊ
*/
static char *bin2hex(char *hex, const void *mem, off_t offset, off_t len)
{
	Intelhex *h = (Intelhex *)hex;
	int sum;
	unsigned char *p, *r;
	char *q;

	if(len <= 0) {
		/* GhR[h */
		strcpy(hex, ":00000001FF");
		return hex;
	}

	/* R[hJnL */
	h->colon = ':';

	/* f[^ */
	int2c(h->length, len);
	sum = (len & 0xff);
	
	/* ItZbgAhX */
	int2c(&h->offset[0], (offset & 0xff00) >> 8);
	int2c(&h->offset[2], offset & 0x00ff);
	sum += ((offset & 0xff00) >> 8) + (offset & 0x00ff);
	
	/* R[h^Cv */
	memcpy(h->record_type, "00", 2);
	sum += 0;
	
	/* f[^ */
	r = (unsigned char *)mem + offset;
	for(p = r, q = h->data; p != r + len; p++, q += 2) {
		int2c(q, *p);
		sum += *p;
	}

	/* `FbNT */
	int2c(q, (~sum + 1) & 0xff);
	q += 2;
	*q = 0;

	return hex;
}

/*
	IntelHEX`t@Cǂݍ
*/
size_t readHex(const char *path, void *mem, off_t *ret_begin, size_t mem_size, int check)
{
	FILE *fp;
	size_t size = 0;
	off_t offset, begin = 0xffff, len;
	char buf[256];

	if(path == NULL || (fp = fopen(path, "r")) == NULL)
		return 0;
	while(!feof(fp)) {
		fgets(buf, sizeof(buf), fp);
		len = hex2bin(mem, &offset, buf, mem_size, check);
		if(len < 0 || len == sizeof(buf)) {
			/* sȃR[h */
			fclose(fp);
			return 0;
		} else if (len == 0) {
			/* GhR[h */
			break;
		} else {
			/* f[^R[h */
			size += len;
			begin = (begin < offset ? begin: offset);
		}
	}
	fclose(fp);

	if(ret_begin != NULL)
		*ret_begin = begin;
	return size;
}

/*
	IntelHEX`t@Cǂݍ(t@C̃AhX𖳎)
*/
size_t readHexAbs(const char *path, void *mem, off_t *ret_begin, size_t mem_size, int check)
{
	size_t size;
	off_t begin;

	if((size = readHex(path, NULL, &begin, 0xffff, check)) == 0)
		return 0;
	if(ret_begin != NULL)
		*ret_begin = begin;
	return readHex(path, (mem != NULL ? (unsigned char *)mem - begin: NULL), NULL, mem_size + begin, check);
}

/*
	IntelHEX`t@C
*/
size_t writeHex(const char *path, const void *mem, off_t begin, size_t size)
{
	FILE *fp;
	off_t offset, len;
	char buf[256];
	
	if(path == NULL || (fp = fopen(path, "w")) == NULL)
		return 0;
	for(offset = begin; offset < begin + size; offset += 16) {
		if((len = begin + size - offset) > 16)
			len = 16;
		bin2hex(buf, mem, offset, len);
		strcat(buf, "\n");
		fputs(buf, fp);
	}
	bin2hex(buf, NULL, 0, 0);
	fputs(buf, fp);
	fclose(fp);

	return size;
}

/*
	IntelHEX`t@C(̐擪珑)
*/
size_t writeHexAbs(const char *path, const void *mem, off_t begin, size_t size)
{
	return writeHex(path, (unsigned char *)mem - begin, begin, size);
}

/*
	Copyright 2005, 2006 maruhiro
	All rights reserved. 

	Redistribution and use in source and binary forms, 
	with or without modification, are permitted provided that 
	the following conditions are met: 

	 1. Redistributions of source code must retain the above copyright notice, 
	    this list of conditions and the following disclaimer. 

	 2. Redistributions in binary form must reproduce the above copyright notice, 
	    this list of conditions and the following disclaimer in the documentation 
	    and/or other materials provided with the distribution. 

	THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
	FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
	THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
	OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/* eof */
